@@ -3,17 +3,30 @@ try:
 except ImportError:
     # python2
     pass
+import inspect
+import itertools
+import logging
+
+from collections import OrderedDict
+from collections import defaultdict
+from collections import deque
 from functools import partial
-from collections import defaultdict, OrderedDict
+from six import string_types
+logger = logging.getLogger(__name__)
+logger.addHandler(logging.NullHandler())
 
 
 def listify(obj):
-    return obj if isinstance(obj, (list, type(None))) else [obj]
+    if obj is None:
+        return []
+    else:
+        return obj if isinstance(obj, (list, tuple, type(None))) else [obj]
 
 
 class State(object):
 
-    def __init__(self, name, on_enter=None, on_exit=None):
+    def __init__(self, name, on_enter=None, on_exit=None,
+                 ignore_invalid_triggers=False):
         """
         Args:
             name (string): The name of the state
@@ -23,22 +36,26 @@ class State(object):
             on_exit (string, list): Optional callable(s) to trigger when a
                 state is exited. Can be either a string providing the name of a
                 callable, or a list of strings.
+            ignore_invalid_triggers (Boolean): Optional flag to indicate if
+                unhandled/invalid triggers should raise an exception
+
         """
         self.name = name
+        self.ignore_invalid_triggers = ignore_invalid_triggers
         self.on_enter = listify(on_enter) if on_enter else []
         self.on_exit = listify(on_exit) if on_exit else []
 
     def enter(self, event_data):
         """ Triggered when a state is entered. """
         for oe in self.on_enter:
-            event_data.machine.callback(
-                getattr(event_data.model, oe), event_data)
+            event_data.machine._callback(oe, event_data)
+        logger.info("%sEntered state %s", event_data.machine.id, self.name)
 
     def exit(self, event_data):
         """ Triggered when a state is exited. """
         for oe in self.on_exit:
-            event_data.machine.callback(
-                getattr(event_data.model, oe), event_data)
+            event_data.machine._callback(oe, event_data)
+        logger.info("%sExited state %s", event_data.machine.id, self.name)
 
     def add_callback(self, trigger, func):
         """ Add a new enter or exit callback.
@@ -51,23 +68,45 @@ class State(object):
         callback_list.append(func)
 
 
-class Transition(object):
-
-    class Condition(object):
+class Condition(object):
 
         def __init__(self, func, target=True):
+            """
+            Args:
+                func (string): Name of the condition-checking callable
+                target (bool): Indicates the target state--i.e., when True,
+                    the condition-checking callback should return True to pass,
+                    and when False, the callback should return False to pass.
+            Notes:
+                This class should not be initialized or called from outside a
+                Transition instance, and exists at module level (rather than
+                nesting under the ransition class) only because of a bug in
+                dill that prevents serialization under Python 2.7.
+            """
             self.func = func
             self.target = target
 
-        def check(self, model):
+        def check(self, event_data):
             """ Check whether the condition passes.
             Args:
-                model (object): the data model attached to the current Machine.
+                event_data (EventData): An EventData instance to pass to the
+                condition (if event sending is enabled) or to extract arguments
+                from (if event sending is disabled). Also contains the data
+                model attached to the current machine which is used to invoke
+                the condition.
             """
-            return getattr(model, self.func)() == self.target
+            predicate = getattr(event_data.model, self.func)
+            if event_data.machine.send_event:
+                return predicate(event_data) == self.target
+            else:
+                return predicate(
+                    *event_data.args, **event_data.kwargs) == self.target
+
+
+class Transition(object):
 
     def __init__(self, source, dest, conditions=None, unless=None, before=None,
-                 after=None):
+                 after=None, prepare=None):
         """
         Args:
             source (string): The name of the source State.
@@ -82,45 +121,64 @@ class Transition(object):
             before (string or list): callbacks to trigger before the
                 transition.
             after (string or list): callbacks to trigger after the transition.
+            prepare (string or list): callbacks to trigger before conditions are checked
         """
         self.source = source
         self.dest = dest
+        self.prepare = [] if prepare is None else listify(prepare)
         self.before = [] if before is None else listify(before)
         self.after = [] if after is None else listify(after)
 
         self.conditions = []
         if conditions is not None:
             for c in listify(conditions):
-                self.conditions.append(self.Condition(c))
+                self.conditions.append(Condition(c))
         if unless is not None:
             for u in listify(unless):
-                self.conditions.append(self.Condition(u, target=False))
+                self.conditions.append(Condition(u, target=False))
 
     def execute(self, event_data):
         """ Execute the transition.
         Args:
             event: An instance of class EventData.
+        Returns: boolean indicating whether or not the transition was
+            successfully executed (True if successful, False if not).
         """
+        logger.info("%sInitiating transition from state %s to state %s...",
+                    event_data.machine.id, self.source, self.dest)
         machine = event_data.machine
-        for c in self.conditions:
-            if not c.check(event_data.model):
-                return False
 
+        for func in self.prepare:
+            machine._callback(getattr(event_data.model, func), event_data)
+            logger.info("Executing callback '%s' before conditions." % func)
+
+        for c in self.conditions:
+            if not c.check(event_data):
+                logger.info("%sTransition condition failed: %s() does not " +
+                            "return %s. Transition halted.", event_data.machine.id, c.func, c.target)
+                return False
         for func in self.before:
-            machine.callback(getattr(event_data.model, func), event_data)
-        machine.get_state(self.source).exit(event_data)
-        machine.set_state(self.dest)
-        event_data.update()
-        machine.get_state(self.dest).enter(event_data)
+            machine._callback(func, event_data)
+            logger.info("%sExecuting callback '%s' before transition.", event_data.machine.id, func)
+
+        self._change_state(event_data)
+
         for func in self.after:
-            machine.callback(getattr(event_data.model, func), event_data)
+            machine._callback(func, event_data)
+            logger.info("%sExecuted callback '%s' after transition.", event_data.machine.id, func)
         return True
 
+    def _change_state(self, event_data):
+        event_data.machine.get_state(self.source).exit(event_data)
+        event_data.machine.set_state(self.dest)
+        event_data.update()
+        event_data.machine.get_state(self.dest).enter(event_data)
+
     def add_callback(self, trigger, func):
-        """ Add a new before or after callback.
+        """ Add a new before, after, or prepare callback.
         Args:
             trigger (string): The type of triggering event. Must be one of
-                'before' or 'after'.
+                'before', 'after' or 'prepare'.
             func (string): The name of the callback function.
         """
         callback_list = getattr(self, trigger)
@@ -129,15 +187,17 @@ class Transition(object):
 
 class EventData(object):
 
-    def __init__(self, state, event, machine, model, *args, **kwargs):
+    def __init__(self, state, event, machine, model, args, kwargs):
         """
         Args:
             state (State): The State from which the Event was triggered.
             event (Event): The triggering Event.
             machine (Machine): The current Machine instance.
             model (object): The model/object the machine is bound to.
-            args and kwargs: Optional positional or named arguments that will
-                be stored internally for possible later use.
+            args (list): Optional positional arguments from trigger method
+                to store internally for possible later use.
+            kwargs (dict): Optional keyword arguments from trigger method
+                to store internally for possible later use.
         """
         self.state = state
         self.event = event
@@ -171,35 +231,60 @@ class Event(object):
             transition (Transition): The Transition instance to add to the
                 list.
         """
-        source = transition.source
         self.transitions[transition.source].append(transition)
 
     def trigger(self, *args, **kwargs):
-        """ Serially execute all transitions that match the current state, 
+        f = partial(self._trigger, *args, **kwargs)
+        return self.machine._process(f)
+
+    def _trigger(self, *args, **kwargs):
+        """ Serially execute all transitions that match the current state,
         halting as soon as one successfully completes.
         Args:
             args and kwargs: Optional positional or named arguments that will
                 be passed onto the EventData object, enabling arbitrary state
                 information to be passed on to downstream triggered functions.
+        Returns: boolean indicating whether or not a transition was
+            successfully executed (True if successful, False if not).
         """
         state_name = self.machine.current_state.name
         if state_name not in self.transitions:
-            raise MachineError(
-                "Can't trigger event %s from state %s!" % (self.name,
-                                                           state_name))
-        event = EventData(self.machine.current_state, self,
-                          self.machine, self.machine.model, *args, **kwargs)
+            msg = "%sCan't trigger event %s from state %s!" % (self.machine.id, self.name,
+                                                               state_name)
+            if self.machine.current_state.ignore_invalid_triggers:
+                logger.warning(msg)
+                return False
+            else:
+                raise MachineError(msg)
+        event = EventData(self.machine.current_state, self, self.machine,
+                          self.machine.model, args=args, kwargs=kwargs)
         for t in self.transitions[state_name]:
+            event.transition = t
             if t.execute(event):
                 return True
         return False
 
+    def add_callback(self, trigger, func):
+        """ Add a new before or after callback to all available transitions.
+        Args:
+            trigger (string): The type of triggering event. Must be one of
+                'before', 'after' or 'prepare'.
+            func (string): The name of the callback function.
+        """
+        for t in itertools.chain(*self.transitions.values()):
+            t.add_callback(trigger, func)
+
 
 class Machine(object):
 
+    # Callback naming parameters
+    callbacks = ['before', 'after', 'prepare', 'on_enter', 'on_exit']
+    separator = '_'
+
     def __init__(self, model=None, states=None, initial=None, transitions=None,
                  send_event=False, auto_transitions=True,
-                 ordered_transitions=False):
+                 ordered_transitions=False, ignore_invalid_triggers=None,
+                 before_state_change=None, after_state_change=None, name=None, queued=False, **kwargs):
         """
         Args:
             model (object): The object whose states we want to manage. If None,
@@ -223,14 +308,42 @@ class Machine(object):
             ordered_transitions (boolean): Convenience argument that calls
                 add_ordered_transitions() at the end of initialization if set
                 to True.
+            ignore_invalid_triggers: when True, any calls to trigger methods
+                that are not valid for the present state (e.g., calling an
+                a_to_b() trigger when the current state is c) will be silently
+                ignored rather than raising an invalid transition exception.
+            before_state_change: A callable called on every change state before
+                the transition happened. It receives the very same args as normal
+                callbacks
+            after_state_change: A callable called on every change state after
+                the transition happened. It receives the very same args as normal
+                callbacks
+            name: If a name is set, it will be used as a prefix for logger output
+            queued (boolean): When True, processes transitions sequentially. A trigger
+                executed in a state callback function will be queued and executed later.
+                Due to the nature of the queued processing, all transitions will
+                _always_ return True since conditional checks cannot be conducted at queueing time.
 
+            **kwargs additional arguments passed to next class in MRO. This can be ignored in most cases.
         """
+
+        try:
+            super(Machine, self).__init__(**kwargs)
+        except TypeError as e:
+            raise MachineError('Passing arguments {0} caused an inheritance error: {1}'.format(kwargs.keys(), e))
+
         self.model = self if model is None else model
         self.states = OrderedDict()
         self.events = {}
         self.current_state = None
         self.send_event = send_event
         self.auto_transitions = auto_transitions
+        self.ignore_invalid_triggers = ignore_invalid_triggers
+        self.before_state_change = before_state_change
+        self.after_state_change = after_state_change
+        self.id = name + ": " if name is not None else ""
+        self._queued = queued
+        self._transition_queue = deque()
 
         if initial is None:
             self.add_states('initial')
@@ -245,11 +358,32 @@ class Machine(object):
         if transitions is not None:
             transitions = listify(transitions)
             for t in transitions:
-                self.add_transition(**t)
+                if isinstance(t, list):
+                    self.add_transition(*t)
+                else:
+                    self.add_transition(**t)
 
         if ordered_transitions:
             self.add_ordered_transitions()
 
+    @staticmethod
+    def _create_transition(*args, **kwargs):
+        return Transition(*args, **kwargs)
+
+    @staticmethod
+    def _create_event(*args, **kwargs):
+        return Event(*args, **kwargs)
+
+    @property
+    def initial(self):
+        """ Return the initial state. """
+        return self._initial
+
+    @property
+    def has_queue(self):
+        """ Return boolean indicating if machine has queue or not """
+        return self._queued
+
     def is_state(self, state):
         """ Check whether the current state matches the named state. """
         return self.current_state.name == state
@@ -262,7 +396,7 @@ class Machine(object):
 
     def set_state(self, state):
         """ Set the current state. """
-        if isinstance(state, str):
+        if isinstance(state, string_types):
             state = self.get_state(state)
         self.current_state = state
         self.model.state = self.current_state.name
@@ -271,7 +405,8 @@ class Machine(object):
         """ Alias for add_states. """
         self.add_states(*args, **kwargs)
 
-    def add_states(self, states, on_enter=None, on_exit=None):
+    def add_states(self, states, on_enter=None, on_exit=None,
+                   ignore_invalid_triggers=None):
         """ Add new state(s).
         Args:
             state (list, string, dict, or State): a list, a State instance, the
@@ -282,23 +417,50 @@ class Machine(object):
                 entered. Only valid if first argument is string.
             on_exit (string or list): callbacks to trigger when the state is
                 exited. Only valid if first argument is string.
+            ignore_invalid_triggers: when True, any calls to trigger methods
+                that are not valid for the present state (e.g., calling an
+                a_to_b() trigger when the current state is c) will be silently
+                ignored rather than raising an invalid transition exception.
+                Note that this argument takes precedence over the same
+                argument defined at the Machine level, and is in turn
+                overridden by any ignore_invalid_triggers explicitly
+                passed in an individual state's initialization arguments.
         """
+
+        ignore = ignore_invalid_triggers
+        if ignore is None:
+            ignore = self.ignore_invalid_triggers
+
         states = listify(states)
         for state in states:
-            if isinstance(state, str):
-                state = State(state, on_enter=on_enter, on_exit=on_exit)
+            if isinstance(state, string_types):
+                state = State(
+                    state, on_enter=on_enter, on_exit=on_exit,
+                    ignore_invalid_triggers=ignore)
             elif isinstance(state, dict):
+                if 'ignore_invalid_triggers' not in state:
+                    state['ignore_invalid_triggers'] = ignore
                 state = State(**state)
             self.states[state.name] = state
             setattr(self.model, 'is_%s' %
                     state.name, partial(self.is_state, state.name))
+            state_name = state.name
+            #  Add enter/exit callbacks if there are existing bound methods
+            enter_callback = 'on_enter_' + state_name
+            if hasattr(self.model, enter_callback) and \
+                    inspect.ismethod(getattr(self.model, enter_callback)):
+                state.add_callback('enter', enter_callback)
+            exit_callback = 'on_exit_' + state_name
+            if hasattr(self.model, exit_callback) and \
+                    inspect.ismethod(getattr(self.model, exit_callback)):
+                state.add_callback('exit', exit_callback)
         # Add automatic transitions after all states have been created
         if self.auto_transitions:
-            for s in states:
+            for s in self.states.keys():
                 self.add_transition('to_%s' % s, '*', s)
 
     def add_transition(self, trigger, source, dest, conditions=None,
-                       unless=None, before=None, after=None):
+                       unless=None, before=None, after=None, prepare=None, **kwargs):
         """ Create a new Transition instance and add it to the internal list.
         Args:
             trigger (string): The name of the method that will trigger the
@@ -318,17 +480,25 @@ class Machine(object):
                 otherwise.
             before (string or list): Callables to call before the transition.
             after (string or list): Callables to call after the transition.
-
+            prepare (string or list): Callables to call when the trigger is activated
+            **kwargs: Additional arguments which can be passed to the created transition.
+                This is useful if you plan to extend Machine.Transition and require more parameters.
         """
         if trigger not in self.events:
-            self.events[trigger] = Event(trigger, self)
+            self.events[trigger] = self._create_event(trigger, self)
             setattr(self.model, trigger, self.events[trigger].trigger)
 
-        if isinstance(source, str):
+        if isinstance(source, string_types):
             source = list(self.states.keys()) if source == '*' else [source]
 
+        if self.before_state_change:
+            before = listify(before) + listify(self.before_state_change)
+
+        if self.after_state_change:
+            after = listify(after) + listify(self.after_state_change)
+
         for s in source:
-            t = Transition(s, dest, conditions, unless, before, after)
+            t = self._create_transition(s, dest, conditions, unless, before, after, prepare, **kwargs)
             self.events[trigger].add_transition(t)
 
     def add_ordered_transitions(self, states=None, trigger='next_state',
@@ -359,7 +529,7 @@ class Machine(object):
                 states.remove(self._initial)
             self.add_transition(trigger, states[-1], states[0])
 
-    def callback(self, func, event_data):
+    def _callback(self, func, event_data):
         """ Trigger a callback function, possibly wrapping it in an EventData
         instance.
         Args:
@@ -368,22 +538,80 @@ class Machine(object):
                 callback (if event sending is enabled) or to extract arguments
                 from (if event sending is disabled).
         """
+        if isinstance(func, string_types):
+            func = getattr(event_data.model, func)
+
         if self.send_event:
             func(event_data)
         else:
             func(*event_data.args, **event_data.kwargs)
 
-    def __getattr__(self, name):
-        terms = name.split('_')
-        if terms[0] in ['before', 'after']:
-            name = '_'.join(terms[1:])
-            if name not in self.events:
-                raise MachineError('Event "%s" is not registered.' % name)
-            return partial(self.events[name].add_callback, terms[0])
+    def _process(self, trigger):
 
-        elif name.startswith('on_enter') or name.startswith('on_exit'):
-            state = self.get_state('_'.join(terms[2:]))
-            return partial(state.add_callback, terms[1])
+        # default processing
+        if not self.has_queue:
+            if not self._transition_queue:
+                # if trigger raises an Error, it has to be handled by the Machine.process caller
+                return trigger()
+            else:
+                raise MachineError("Attempt to process events synchronously while transition queue is not empty!")
+
+        # process queued events
+        self._transition_queue.append(trigger)
+        # another entry in the queue implies a running transition; skip immediate execution
+        if len(self._transition_queue) > 1:
+            return True
+
+        # execute as long as transition queue is not empty
+        while self._transition_queue:
+            try:
+                self._transition_queue[0]()
+                self._transition_queue.popleft()
+            except Exception:
+                # if a transition raises an exception, clear queue and delegate exception handling
+                self._transition_queue.clear()
+                raise
+        return True
+
+    @classmethod
+    def _identify_callback(cls, name):
+        # Does the prefix match a known callback?
+        try:
+            callback_type = cls.callbacks[[name.find(x) for x in cls.callbacks].index(0)]
+        except ValueError:
+            return None, None
+
+        # Extract the target by cutting the string after the type and separator
+        target = name[len(callback_type) + len(cls.separator):]
+
+        # Make sure there is actually a target to avoid index error and enforce _ as a separator
+        if target == '' or name[len(callback_type)] != cls.separator:
+            return None, None
+
+        return callback_type, target
+
+    def __getattr__(self, name):
+        if name.startswith('__'):
+            if name in self.__dict__:
+                return self.__dict__[name]
+            else:
+                raise AttributeError("{} does not exist".format(name))
+
+        # Could be a callback
+        callback_type, target = self._identify_callback(name)
+
+        if callback_type is not None:
+            if callback_type in ['before', 'after', 'prepare']:
+                if target not in self.events:
+                    raise MachineError('Event "%s" is not registered.' % target)
+                return partial(self.events[target].add_callback, callback_type)
+
+            elif callback_type in ['on_enter', 'on_exit']:
+                state = self.get_state(target)
+                return partial(state.add_callback, callback_type[3:])
+
+        # Nothing matched
+        raise AttributeError("{} does not exist".format(name))
 
 
 class MachineError(Exception):
